import unittest

from superspreader.exceptions import ValidationException
from superspreader.fields import BaseField, TimecodeField


class DummyField(BaseField):
    target_type = int


class BaseFieldTestCase(unittest.TestCase):
    def test_required(self):
        test_field = DummyField(source="test")

        with self.assertRaises(ValidationException) as cm:
            test_field(None, "en")

        self.assertEqual(cm.exception.msg, "“test” is required")

        test_field = DummyField(source="test", required=False)
        self.assertEqual(test_field(None, language="en"), None)

        # Reverse case.
        value = test_field(666, "en")
        self.assertEqual(value, 666)

    def test_wrong_type(self):
        test_field = DummyField(source="test")

        with self.assertRaises(ValidationException) as cm:
            test_field("I’m a string", "en")

        self.assertEqual(
            cm.exception.msg, "“test” should be of type int, but it’s of type str"
        )

    def test_timecode_field(self):
        test_field = TimecodeField(source="test")
        timecode_str = "00:13:06,9"
        invalid_timecode_str = "asdf123"

        value = test_field(timecode_str, language="en")
        self.assertEqual(value, 786.9)

        with self.assertRaises(ValidationException) as cm:
            test_field(invalid_timecode_str, language="en")

        self.assertEqual(cm.exception.msg, "“test” has an invalid format: asdf123")



import datetime
import re
from abc import ABC

from .exceptions import ImproperlyConfigured, ValidationException
from .i18n import translate as _


class BaseField(ABC):
    target_type = None
    default = None
    cast_type = True

    def __init__(self, source, required=True, default=None, unique=False):
        """
        :param str source: The column name from the spreadsheet
        :param bool required: Indicates whether a non-`None` value is required.
                              Ignored when a default value is provided.
        :param any default: The default value to use
        :param bool unique: Indicates whether values must be unique
        """
        self.source = source
        self.required = required
        self.unique = unique

        self.language = None
        self.extra_context = None

        if default is not None:
            self.default = default

    def __call__(self, value, language, extra_context=None):
        if extra_context is None:
            extra_context = {}

        self.extra_context = extra_context
        self.language = language

        if value is None:
            value = self.get_default()

        if value is None and self.required is True:
            msg = _("field.is_required", params={"field": self.source})
            raise ValidationException(msg)

        return self.clean(value)

    def get_target_type(self):
        """
        The field value’s desired type, used for casting primitive types.
        In more advanced cases, disable type casting (cast_type=False) and implement
        your own logic in `clean`.
        """
        return self.target_type

    def get_default(self):
        """
        Gets the default value for the field
        """
        return self.default

    def clean(self, value):
        if value is None:
            return

        if self.cast_type:
            desired_type = self.get_target_type()
            if desired_type is None:
                raise ImproperlyConfigured(
                    f"You have to specify a target type for"
                    f" {self.__class__.__name__}"
                )

            if not isinstance(value, desired_type):
                try:
                    value = desired_type(value)
                except (
                    TypeError,
                    ValueError,
                ):
                    params = {
                        "field": self.source,
                        "target_type": self.get_target_type().__name__,
                        "actual_type": value.__class__.__name__,
                    }
                    msg = _("field.wrong_type", params=params)
                    raise ValidationException(msg)

        return value


class CharField(BaseField):
    target_type = str


class DateField(BaseField):
    target_type = datetime.date

    def clean(self, value):
        if isinstance(value, datetime.datetime):
            value = value.date()

        return super().clean(value)


class DateTimeField(BaseField):
    target_type = datetime.datetime


class FloatField(BaseField):
    target_type = float


class IntegerField(BaseField):
    target_type = int


class TimecodeField(FloatField):
    """
    Parse _timecode values like 00:13:06,9 to float value. This is commonly
    used in video and audio processing.
    """

    def clean(self, value):
        if isinstance(value, str):
            value = self._timecode(value)

        return super().clean(value)

    def _timecode(self, timecode_string):
        timecode_string = timecode_string.strip()
        timecode_regex = re.compile(r"^(\d{1,2}):(\d{1,2}):(\d{1,2})([-,](\d))?$")
        matches = timecode_regex.match(timecode_string)

        try:
            hours = int(matches.group(1))
            minutes = int(matches.group(2))
            seconds = int(matches.group(3))

            millis = matches.group(5)
            if millis:
                millis = int(millis)
            else:
                millis = 0
        except Exception:
            msg = _(
                "field.timecode_parse_error",
                params={"_timecode": timecode_string, "field": self.source},
            )
            raise ValidationException(msg)

        seconds = hours * 3600 + minutes * 60 + seconds + millis / 10

        return seconds



# the inclusion of the tests module is not meant to offer best practices for
# testing in general, but rather to support the `find_packages` example in
# setup.py that excludes installing the "tests" package

import datetime
import unittest
from unittest.mock import MagicMock

from superspreader import fields
from superspreader.sheets import BaseSheet
from tests import file_path


class AlbumSheet(BaseSheet):
    sheet_name = "Albums"
    header_rows = 3
    label_row = 2

    artist = fields.CharField(source="Artist")
    album = fields.CharField(source="Album")
    release_date = fields.DateField(source="Release Date")
    average_review = fields.FloatField(source="Average Review")
    chart_position = fields.IntegerField(source="Chart Position")


class TestFullImport(unittest.TestCase):
    """
    Test the process of importing a spreadsheet
    """

    def setUp(self) -> None:
        self.sheet = AlbumSheet(file_path("albums.xlsx"))

    def test_basics(self):
        self.sheet.load()
        self.assertFalse(self.sheet.has_errors)
        self.assertEqual(len(self.sheet), 3)

    def test_record(self):
        self.sheet.load()
        first_record = self.sheet[0]

        self.assertEqual(first_record.get("album"), "Toy")
        self.assertEqual(first_record.get("artist"), "David Bowie")
        self.assertEqual(first_record.get("release_date"), datetime.date(2022, 1, 7))
        self.assertEqual(first_record.get("average_review"), 4.3)
        self.assertEqual(first_record.get("chart_position"), 5)

    def test_rows(self):
        self.sheet.load()
        rows = self.sheet.rows(exclude=["album"])
        with self.assertRaises(KeyError):
            rows[0]["album"]

    def test_info(self):
        self.sheet.load()
        self.assertEqual("Row 6: Skipped row", self.sheet.infos[0])

    def test_errors(self):
        path = file_path("albums_with_errors.xlsx")
        sheet_with_errors = AlbumSheet(path)
        sheet_with_errors.load()

        self.assertTrue(sheet_with_errors.has_errors)
        self.assertEqual(sheet_with_errors.errors[0], "Row 7: “Album” is required")

    def test_empty_sheet_with_extra_data(self):
        """Tests whether empty rows are skipped, even when extra data is provided"""
        path = file_path("albums_empty.xlsx")
        sheet = AlbumSheet(path, extra_data={"status": "released"})
        self.assertFalse(sheet.has_errors)
        self.assertEqual(len(sheet), 0)

    def test_extra_data_static(self):
        """Tests whether extra data is returned in the resulting row"""
        fp = file_path("albums.xlsx")
        sheet = AlbumSheet(path=fp, extra_data={"status": "released"})
        sheet.load()
        self.assertEqual(sheet.rows()[0].get("status"), "released")

    def test_extra_data_dynamic(self):
        """Tests whether callables in extra data are called"""
        fp = file_path("albums.xlsx")
        test_fn = MagicMock()
        sheet = AlbumSheet(path=fp, extra_data={"test_fn": test_fn})
        sheet.load()
        test_fn.assert_called()

    def test_extra_data_dynamic_args(self):
        """Tests whether callables in extra data are called with the row"""
        fp = file_path("albums.xlsx")

        def test_fn(row):
            self.assertIsInstance(row, dict)

        sheet = AlbumSheet(path=fp, extra_data={"test_fn": test_fn})
        sheet.load()



# the inclusion of the tests module is not meant to offer best practices for
# testing in general, but rather to support the `find_packages` example in
# setup.py that excludes installing the "tests" package

import unittest

from superspreader import fields
from superspreader.exceptions import ImproperlyConfigured
from superspreader.sheets import BaseSheet
from tests import file_path


class ContactSheet(BaseSheet):
    sheet_name = "Contacts"

    id = fields.CharField(source="ID", unique=True)
    name = fields.CharField(source="Name")


class SheetTestCase(unittest.TestCase):
    def test_sheet_name(self):
        class AlbumSheet(BaseSheet):
            pass

        with self.assertRaises(ImproperlyConfigured):
            AlbumSheet(path="test")

    def test_path_or_file(self):
        class AlbumSheet(BaseSheet):
            pass

        with self.assertRaises(ImproperlyConfigured):
            AlbumSheet(path="test")

    def test_no_header_rows(self):
        class AlbumSheet(BaseSheet):
            header_rows = 0

        with self.assertRaises(ImproperlyConfigured):
            AlbumSheet(path="test")

    def test_attributes_are_inherited(self):
        class ASheet(BaseSheet):
            sheet_name = "A Sheet"
            a_field = fields.CharField(source="")

        class BSheet(ASheet):
            pass

        sheet = BSheet(path="test")
        sheet_fields = sheet._build_fields()

        self.assertTrue("a_field" in sheet_fields)

    def test_unique_violation(self):
        """
        Tests whether unique fields are validate properly
        """
        path = file_path("contacts_duplicates.xlsx")
        sheet = ContactSheet(path)
        sheet.load()

        self.assertEqual(len(sheet.errors), 1)
        self.assertEqual(
            sheet.errors[0],
            "“ID” must contain unique values only, but “1” occurs 1 times",
        )



import pathlib

from setuptools import find_packages, setup

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="superspreader",
    version="0.2.7",
    description="Load data from spreadsheets easily",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/julianklotz/superspreader",
    author="Julian Klotz",
    author_email="post@julianklotz.de",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords="excel, spreadsheets, import, csv, tsv, openpyxl",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.6, <4",
    install_requires=["openpyxl>=3"],
    extras_require={
        "dev": ["pre-commit"],
        "test": ["coverage"],
    },
    project_urls={
        "Source": "https://github.com/julianklotz/superspreader",
    },
)



from .i18n import DE, EN

messages = {
    EN: {
        "field.wrong_type": (
            "“%(field)s” should be of type %(target_type)s, but it’s of type %(actual_type)s",
            None,
        ),
        "field.is_required": ("“%(field)s” is required", None),
        "field.timecode_parse_error": (
            "“%(field)s” has an invalid format: %(_timecode)s",
            None,
        ),
        "field.related_does_not_exist": (
            "Related object identified by %(value)s on “%(field)s” does not exist",
            None,
        ),
        "field.related_multiple_objects_returned": (
            "More than one related object identified by %(value)s on “%(field)s”",
            None,
        ),
        "sheet.row_info": ("Row %(row)s: %(message)s", None),
        "sheet.column_missing": ("Column “%(column)s” not present in sheet", None),
        "sheet.sheet_missing": ("Sheet “%(sheet)s” not present in document", None),
        "sheet.unique_violation": (
            "“%(column)s” must contain unique values only, but “%(value)s” occurs %(total)i times",
            None,
        ),
    },
    DE: {
        "field.wrong_type": (
            "„%(field)s“ sollte den Typ %(target_type)s haben, aber der"
            "Typ ist %(actual_type)s",
            None,
        ),
        "field.is_required": ("„%(field)s“ muss ausgefüllt sein", None),
        "field.timecode_parse_error": (
            "„%(field)s“ hat ein ungültiges Format: %(_timecode)s",
            None,
        ),
        "field.related_does_not_exist": (
            "Ein Objekt mit %(value)s bei „%(field)s“ existiert nicht",
            None,
        ),
        "field.related_multiple_objects_returned": (
            "Mehr als ein Objekt mit %(value)s beim „%(field)s“ gefunden",
            None,
        ),
        "sheet.row_info": ("Zeile %(row)s: %(message)s", None),
        "sheet.column_missing": ("Die Spalte „%(column)s“ fehlt im Blatt", None),
        "sheet.sheet_missing": ("Das Blatt „%(sheet)s“ ist nicht vorhanden", None),
        "sheet.unique_violation": (
            "„%(column)s“ darf nur eindeutige Werte enthalten, „%(value)s“ kommt aber %(total)i mal vor",
            None,
        ),
    },
}
